import colorsys
import logging
import math
import sys
from collections import deque

import numpy as np
import numpy.typing as npt
import pygame
import yaml
from scipy.spatial import KDTree

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.game_items import (
    ProgressibleItem,
    Item,
    CookingEquipment,
    Meal,
    Plate,
)
from overcooked_simulator.overcooked_environment import Action
from overcooked_simulator.pygame_gui.game_colors import BLUE
from overcooked_simulator.pygame_gui.game_colors import colors, Color
from overcooked_simulator.simulation_runner import Simulator

USE_PLAYER_COOK_SPRITES = True
SHOW_INTERACTION_RANGE = False


def create_polygon(n, length):
    if n == 0:
        return np.array([0, 0])

    vector = np.array([length, 0])
    angle = (2 * np.pi) / n

    rot_matrix = np.array(
        [[np.cos(angle), -np.sin(angle)], [np.sin(angle), np.cos(angle)]]
    )

    vecs = [vector]
    for i in range(n - 1):
        vector = np.dot(rot_matrix, vector)
        vecs.append(vector)

    return vecs


log = logging.getLogger(__name__)


class PlayerKeySet:
    """Set of keyboard keys for controlling a player.
    First four keys are for movement. Order: Down, Up, Left, Right.
    5th key is for interacting with counters.
    6th key ist for picking up things or dropping them.

    """

    def __init__(self, player_name: str, keys: list[pygame.key]):
        """Creates a player key set which contains information about which keyboard keys control the player.
        Movement keys in the following order: Down, Up, Left, Right

        Args:
            player_name: The name of the player to control.
            keys: The keys which control this player in the following order: Down, Up, Left, Right, Interact, Pickup.
        """
        self.name = player_name
        self.player_keys = keys
        self.move_vectors = [[-1, 0], [1, 0], [0, -1], [0, 1]]
        self.key_to_movement = {
            key: vec for (key, vec) in zip(self.player_keys[:-2], self.move_vectors)
        }
        self.interact_key = self.player_keys[-2]
        self.pickup_key = self.player_keys[-1]


class PyGameGUI:
    """Visualisation of the overcooked environment and reading keyboard inputs using pygame."""

    def __init__(
        self,
        simulator: Simulator,
        player_names: list[str],
        player_keys: list[pygame.key],
    ):
        self.screen = None
        self.FPS = 60
        self.simulator = simulator
        self.counter_size = self.simulator.env.counter_side_length
        self.window_width, self.window_height = (
            simulator.env.world_width,
            simulator.env.world_height,
        )

        self.player_names = player_names
        self.player_keys = player_keys

        self.player_key_sets: list[PlayerKeySet] = [
            PlayerKeySet(player_name, keys)
            for player_name, keys in zip(
                self.player_names, self.player_keys[: len(self.player_names)]
            )
        ]

        # TODO cache loaded images?
        with open(ROOT_DIR / "pygame_gui" / "visualization.yaml", "r") as file:
            self.visualization_config = yaml.safe_load(file)

        self.images_path = ROOT_DIR / "pygame_gui" / "images"

        self.player_colors = self.create_player_colors()

        self.image_cache_dict = {}

    def create_player_colors(self) -> list[Color]:
        number_player = len(self.simulator.env.players)
        hue_values = np.linspace(0, 1, number_player + 1)

        colors_vec = np.array([col for col in colors.values()])

        tree = KDTree(colors_vec)

        color_names = list(colors.keys())

        player_colors = []
        for hue in hue_values:
            rgb = colorsys.hsv_to_rgb(hue, 1, 1)
            query_color = np.array([int(c * 255) for c in rgb])
            _, index = tree.query(query_color, k=1)
            player_colors.append(color_names[index])

        return player_colors

    def send_action(self, action: Action):
        """Sends an action to the game environment.

        Args:
            action: The action to be sent. Contains the player, action type and move direction if action is a movement.
        """
        self.simulator.enter_action(action)

    def handle_keys(self):
        """Handles keyboard inputs. Sends action for the respective players. When a key is held down, every frame
        an action is sent in this function.
        """
        keys = pygame.key.get_pressed()
        for player_idx, key_set in enumerate(self.player_key_sets):
            relevant_keys = [keys[k] for k in key_set.player_keys]
            if any(relevant_keys[:-2]):
                move_vec = np.zeros(2)
                for idx, pressed in enumerate(relevant_keys[:-2]):
                    if pressed:
                        move_vec += key_set.move_vectors[idx]
                if np.linalg.norm(move_vec) != 0:
                    move_vec = move_vec / np.linalg.norm(move_vec)

                action = Action(key_set.name, "movement", move_vec)
                self.send_action(action)

    def handle_key_event(self, event):
        """Handles key events for the pickup and interaction keys. Pickup is a single action,
        for interaction keydown and keyup is necessary, because the player has to be able to hold
        the key down.

        Args:
            event: Pygame event for extracting the key action.
        """
        for key_set in self.player_key_sets:
            if event.key == key_set.pickup_key and event.type == pygame.KEYDOWN:
                action = Action(key_set.name, "pickup", "pickup")
                self.send_action(action)

            if event.key == key_set.interact_key:
                if event.type == pygame.KEYDOWN:
                    action = Action(key_set.name, "interact", "keydown")
                    self.send_action(action)
                elif event.type == pygame.KEYUP:
                    action = Action(key_set.name, "interact", "keyup")
                    self.send_action(action)

    def draw_background(self):
        """Visualizes a game background."""
        block_size = self.counter_size // 2  # Set the size of the grid block
        for x in range(0, self.window_width, block_size):
            for y in range(0, self.window_height, block_size):
                rect = pygame.Rect(x, y, block_size, block_size)
                pygame.draw.rect(
                    self.screen,
                    self.visualization_config["Kitchen"]["background_lines"],
                    rect,
                    1,
                )

    def draw_image(self, img_path, size, pos, rot_angle=0):
        cache_entry = f"{img_path}"
        if cache_entry in self.image_cache_dict.keys():
            image = self.image_cache_dict[cache_entry]
        else:
            image = pygame.image.load(
                ROOT_DIR / "pygame_gui" / img_path
            ).convert_alpha()
            self.image_cache_dict[cache_entry] = image

        image = pygame.transform.scale(image, (size, size))
        if rot_angle != 0:
            image = pygame.transform.rotate(image, rot_angle)
        rect = image.get_rect()
        rect.center = pos
        self.screen.blit(image, rect)

    def draw_players(self, state):
        """Visualizes the players as circles with a triangle for the facing direction.
        If the player holds something in their hands, it is displayed

        Args:
            state: The game state returned by the environment.
        """
        for p_idx, player in enumerate(state["players"].values()):
            if USE_PLAYER_COOK_SPRITES:
                img_path = self.visualization_config["Cook"]["parts"][0]["path"]
                rel_x, rel_y = player.facing_direction
                angle = -np.rad2deg(math.atan2(rel_y, rel_x)) + 90
                size = (
                    self.visualization_config["Cook"]["parts"][0]["size"]
                    * self.counter_size
                )
                self.draw_image(img_path, size, player.pos, angle)

            else:
                pos = player.pos
                size = player.radius * self.counter_size
                color1 = self.player_colors[p_idx]
                color2 = colors["white"]

                pygame.draw.circle(self.screen, color2, pos, size)
                pygame.draw.circle(self.screen, BLUE, pos, size, width=1)
                pygame.draw.circle(self.screen, colors[color1], pos, size // 2)

                pos = player.pos
                facing = player.facing_direction
                pygame.draw.polygon(
                    self.screen,
                    BLUE,
                    (
                        (
                            pos[0] + (facing[1] * 0.1 * self.counter_size),
                            pos[1] - (facing[0] * 0.1 * self.counter_size),
                        ),
                        (
                            pos[0] - (facing[1] * 0.1 * self.counter_size),
                            pos[1] + (facing[0] * 0.1 * self.counter_size),
                        ),
                        player.pos + (facing * 0.5 * self.counter_size),
                    ),
                )

            if SHOW_INTERACTION_RANGE:
                pygame.draw.circle(
                    self.screen, BLUE, player.pos, player.interaction_range, width=1
                )

            if player.holding is not None:
                holding_item_pos = player.pos + (20 * player.facing_direction)
                self.draw_item(holding_item_pos, player.holding)

    def draw_thing(
        self, pos: npt.NDArray[float], parts: list[dict[str]], scale: float = 1.0
    ):
        """Draws an item, based on its visual parts specified in the visualization config.

        Args:
            pos: Where to draw the item parts.
            parts: The visual parts to draw.
            scale: Rescale the item by this factor.
        """
        for part in parts:
            part_type = part["type"]
            if part_type == "image":
                self.draw_image(
                    parts[0]["path"], parts[0]["size"] * scale * self.counter_size, pos
                )
            elif part_type == "rect":
                height = part["height"] * self.counter_size
                width = part["width"] * self.counter_size
                color = part["color"]
                if "center_offset" in part:
                    dx, dy = np.array(part["center_offset"]) * self.counter_size
                    rect = pygame.Rect(pos[0] + dx, pos[1] + dy, height, width)
                    pygame.draw.rect(self.screen, color, rect)
                else:
                    rect = pygame.Rect(
                        pos[0] - (height / 2),
                        pos[1] - (width / 2),
                        height,
                        width,
                    )
                pygame.draw.rect(self.screen, color, rect)
            elif part_type == "circle":
                radius = part["radius"] * self.counter_size
                color = colors[part["color"]]
                if "center_offset" in part:
                    pygame.draw.circle(
                        self.screen,
                        color,
                        pos + np.array(part["center_offset"]),
                        radius,
                    )
                else:
                    pygame.draw.circle(self.screen, color, pos, radius)

    def draw_item(self, pos: npt.NDArray[float], item: Item, scale: float = 1.0):
        """Visualization of an item at the specified position. On a counter or in the hands of the player.
        The visual composition of the item is read in from visualization.yaml file, where it is specified as
        different parts to be drawn.

        Args:
            pos: The position of the item to draw.
            item: The item do be drawn in the game.
            scale: Rescale the item by this factor.
        """

        if not isinstance(item, Meal):
            if item.name in self.visualization_config:
                self.draw_thing(
                    pos, self.visualization_config[item.name]["parts"], scale=scale
                )

        if isinstance(item, (ProgressibleItem, Plate)) and not item.finished:
            self.draw_progress_bar(pos, item.progressed_steps, item.steps_needed)

        if isinstance(item, CookingEquipment) and item.content:
            self.draw_item(pos, item.content)

        if isinstance(item, Meal):
            if item.finished:
                if item.name in self.visualization_config:
                    self.draw_thing(pos, self.visualization_config[item.name]["parts"])
            else:
                for idx, o in enumerate(item.parts):
                    triangle_offsets = create_polygon(len(item.parts), length=10)
                    self.draw_item(pos + triangle_offsets[idx], o, scale=0.6)

    def draw_progress_bar(self, pos, current, needed):
        """Visualize progress of progressing item as a green bar under the item."""
        if current != 0:
            bar_height = self.counter_size * 0.2
            progress_width = (current / needed) * self.counter_size
            progress_bar = pygame.Rect(
                pos[0] - (self.counter_size / 2),
                pos[1] - (self.counter_size / 2) + self.counter_size - bar_height,
                progress_width,
                bar_height,
            )
            pygame.draw.rect(self.screen, colors["green1"], progress_bar)

    def draw_counter(self, counter):
        """Visualization of a counter at its position. If it is occupied by an item, it is also shown.
        The visual composition of the counter is read in from visualization.yaml file, where it is specified as
        different parts to be drawn.

        Args:
            counter: The counter to visualize.
        """

        self.draw_thing(counter.pos, self.visualization_config["Counter"]["parts"])
        if str(counter) in self.visualization_config:
            self.draw_thing(
                counter.pos, self.visualization_config[str(counter)]["parts"]
            )
        else:
            self.draw_thing(
                counter.pos,
                self.visualization_config[counter.__class__.__name__]["parts"],
            )

        if counter.occupied_by is not None:
            # Multiple plates on plate return:
            if isinstance(counter.occupied_by, (list, deque)):
                with self.simulator.env.lock:
                    for i, o in enumerate(counter.occupied_by):
                        self.draw_item(
                            np.abs([counter.pos[0], counter.pos[1] - (i * 3)]), o
                        )
            # All other items:
            else:
                self.draw_item(counter.pos, counter.occupied_by)

    def draw_counters(self, state):
        """Visualizes the counters in the environment.

        Args:
            state: The game state returned by the environment.
        """
        for counter in state["counters"]:
            self.draw_counter(counter)

    def draw(self, state):
        """Main visualization function.

        Args:
            state: The game state returned by the environment.
        """
        self.screen.fill(
            colors[self.visualization_config["Kitchen"]["ground_tiles_color"]]
        )
        self.draw_background()

        self.draw_counters(state)
        self.draw_players(state)

        pygame.display.flip()

    def start_pygame(self):
        """Starts pygame and the gui loop. Each frame the game state is visualized and keyboard inputs are read."""
        log.debug(f"Starting pygame gui at {self.FPS} fps")
        pygame.init()
        pygame.font.init()

        self.screen = pygame.display.set_mode((self.window_width, self.window_height))
        pygame.display.set_caption("Simple Overcooked Simulator")
        self.screen.fill(
            colors[self.visualization_config["Kitchen"]["ground_tiles_color"]]
        )

        clock = pygame.time.Clock()

        # Game loop
        running = True
        while running:
            try:
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        running = False
                    if event.type in [pygame.KEYDOWN, pygame.KEYUP]:
                        self.handle_key_event(event)

                self.handle_keys()
                clock.tick(self.FPS)
                state = self.simulator.get_state()
                self.draw(state)

            except KeyboardInterrupt:
                pygame.quit()
                self.simulator.stop()
                sys.exit()

        pygame.quit()
